""" metadata.py - Variable Attribute Definitions """

__all__ = ["add_metadata", "nominal_coord_metadata"]

import xarray as xr


def add_metadata(dset):
    """Function to add missing metadata for static/grid dataset

    Parameters
    ----------
    dset : xarray.Dataset
        Input dataset containing grid variables

    Returns
    -------
    xarray.Dataset
    """

    geolon = "geolon"
    geolat = "geolat"
    areacello = "areacello"
    geolon_u = "geolon_u"
    geolat_u = "geolat_u"
    areacello_u = "areacello_u"
    geolon_v = "geolon_v"
    geolat_v = "geolat_v"
    areacello_v = "areacello_v"
    geolon_c = "geolon_c"
    geolat_c = "geolat_c"
    areacello_c = "areacello_c"
    deptho = "deptho"
    wet = "wet"
    wet_u = "wet_u"
    wet_v = "wet_v"

    attributes = {}

    attributes[areacello] = {
        "units": "m2",
        "long_name": "Ocean Grid-Cell Area",
        "standard_name": "cell_area",
    }

    attributes[areacello_u] = attributes[areacello]
    attributes[areacello_v] = attributes[areacello]
    attributes[areacello_c] = attributes[areacello]

    attributes[deptho] = {
        "units": "m",
        "long_name": "Sea Floor Depth",
        "standard_name": "sea_floor_depth_below_geoid",
    }

    attributes[geolat] = {
        "units": "degrees_north",
        "long_name": "Latitude of tracer (T) points",
        "standard_name": "latitude",
        "axis": "Y",
    }

    attributes[geolat_u] = {
        "units": "degrees_north",
        "comment": "Latitude of zonal velocity (Cu) points",
        "standard_name": "latitude",
        "axis": "Y",
    }

    attributes[geolat_v] = {
        "units": "degrees_north",
        "comment": "Latitude of meridional velocity (Cv) points",
        "standard_name": "latitude",
        "axis": "Y",
    }

    attributes[geolat_c] = {
        "units": "degrees_north",
        "comment": "Latitude of corner (Bu) point",
        "standard_name": "latitude",
        "axis": "Y",
    }

    attributes[geolon] = {
        "units": "degrees_east",
        "comment": "Longitude of tracer (T) points",
        "standard_name": "longitude",
        "axis": "X",
    }

    attributes[geolon_u] = {
        "units": "degrees_east",
        "comment": "Longitude of zonal velocity (Cu) points",
        "standard_name": "longitude",
        "axis": "X",
    }

    attributes[geolon_v] = {
        "units": "degrees_east",
        "comment": "Longitude of meridional velocity (Cv) points",
        "standard_name": "longitude",
        "axis": "X",
    }

    attributes[geolon_c] = {
        "units": "degrees_east",
        "comment": "Longitude of corner (Bu) point",
        "standard_name": "longitude",
        "axis": "X",
    }

    attributes[wet] = {
        "standard_name": "sea_binary_mask",
        "long_name": "Sea Binary Mask 1 = sea, 0 = land at tracer points",
        "units": "1",
    }

    attributes[wet_u] = {
        "standard_name": "sea_binary_mask",
        "long_name": "Sea Binary Mask 1 = sea, 0 = land at (Cu) points",
        "units": "1",
    }

    attributes[wet_v] = {
        "standard_name": "sea_binary_mask",
        "long_name": "Sea Binary Mask 1 = sea, 0 = land at (Cv) points",
        "units": "1",
    }

    for var in dset.keys():
        if var in attributes.keys():
            dset[var].attrs = attributes[var]
            if "lon" in var or "lat" in var:
                dset = dset.set_coords(var)

    return dset


def nominal_coord_metadata(xobj):
    """Function to reset the nominal coordinate / index coordinate metadata

    Parameters
    ----------
    xobj : xarray.Dataset or xarray.DataArray
        Input data containing nominal/index coordinate(s)

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Object with corrected metadata
    """

    ds = xr.Dataset({xobj.name: xobj}) if isinstance(xobj, xr.DataArray) else xobj
    assert isinstance(ds, xr.Dataset)

    for var in list(ds.keys()) + list(ds.coords) + list(ds.dims):
        if var == "xh":
            ds[var].attrs = {
                "units": "1",
                "long_name": "h point nominal x-index",
            }
        elif var == "yh":
            ds[var].attrs = {
                "units": "1",
                "long_name": "h point nominal y-index",
            }
        elif var == "xq":
            ds[var].attrs = {
                "units": "1",
                "long_name": "q point nominal x-index",
            }
        elif var == "yq":
            ds[var].attrs = {
                "units": "1",
                "long_name": "q point nominal y-index",
            }

    if isinstance(xobj, xr.DataArray):
        result = ds[xobj.name]
    else:
        result = ds

    return result



""" plot.py - plotting routines """

import warnings
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, BoundaryNorm
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk

from . import VerticalSplitScale

from momgrid.geoslice import geoslice
from momgrid.util import infer_bounds

__all__ = [
    "add_stats_box",
    "generate_cmap_and_norm",
    "round_step",
    "update_stats",
    "compare_2d",
]


def add_stats_box(
    ax,
    xloc,
    yloc,
    label,
    min_val,
    max_val,
    mean_val=None,
    rms_val=None,
    ha="left",
    sig_digits=5,
):
    """Function to add a formatted stats box to an axis

    If both mean_val and rms_val are supplied, a 2x2 box will be returned

    Parameters
    ----------
    ax : matplot.pyplot.axis
        Axis object to add a stats box
    xloc : float
        x-location of box (0.0 to 1.0)
    yloc : float
        y-location of box (0.0 to 1.0)
    label : str
        String title for box
    min_val : float
        Minimum value
    max_val : float
        Maximum value
    mean_val : float, optional
        Mean value, by default None
    rms_val : float, optional
        RMS difference / error value, by default None
    ha : string, optional
        Horizontal alignment, by default "left"
    sig_digits : int, optional
        Number of significant digits to report, by default 5
    """

    # Creating the text string with rounded values

    # Construct the text string based on the supplied values
    # Rounding is to 5 significant digits
    if mean_val is not None and rms_val is not None:
        textstr = (
            f"{label}:\n"
            f"min = {min_val:.{sig_digits}g}   max = {max_val:.{sig_digits}g}\n"
            f"mean = {mean_val:.{sig_digits}g}   RMSDiff = {rms_val:.{sig_digits}g}"
        )

    else:
        textstr = (
            f"{label}:\n"
            f"min = {min_val:.{sig_digits}g}\n"
            f"max = {max_val:.{sig_digits}g}"
        )

    # Define the style properties for the box
    props = dict(
        boxstyle="round,pad=0.5", edgecolor="black", linewidth=1.5, facecolor="white"
    )

    # Add the box and text to the axis object
    ax.text(
        xloc,
        yloc,
        textstr,
        transform=ax.transAxes,
        fontsize=8,
        ha=ha,
        verticalalignment="top",
        bbox=props,
    )


def round_step(step):
    """Function to round step interval to a more intuituve value

    Rounds the step to the nearest value ending in 0, 2, or 5.

    Parameters
    ----------
    step : float
       Value of step interval to be rounded

    Returns
    -------
    float
        Rounded step value
    """

    # Get the order of magnitude
    power = 10 ** np.floor(np.log10(step))

    rounded_step = np.round(step / power, 1) * power

    # Adjust to end in 0, 2, or 5
    last_digit = int(str(rounded_step)[-1])
    if last_digit < 2:
        return np.floor(rounded_step / power) * power
    elif last_digit < 5:
        return np.floor(rounded_step / power) * power + 2 * power / 10
    elif last_digit < 7:
        return np.floor(rounded_step / power) * power + 5 * power / 10
    else:
        return np.floor(rounded_step / power) * power + power


def generate_cmap_and_norm(
    min_val,
    max_val,
    levels,
    cmap_name="viridis",
    white_center=False,
    centered_on_zero=False,
):
    """Function to generate a colormap with specific values

    Generates a discrete colormap and normalization for a specified
    range and number of levels, with an option to center the colormap
    on zero and have equal step sizes ending in 0, 2, or 5.

    Parameters
    ----------
    min_val : float
        Minimum value of the range.
    max_val : float
        Maximum value of the range.
    levels : int
        Number of intervals or levels
        (must be odd if centered_on_zero is True).
    cmap_name : str, optional
        Name of an existing colormap, by default "viridis"
    white_center : bool
        Whether to set the middle interval color to white,
        by default False
    centered_on_zero : bool
        Whether to center the middle interval on zero,
        by default False

    Returns
    -------
    cmap : ListedColormap
    norm : BoundaryNorm instance
    """

    # Check if requested levels is consisted with options
    if centered_on_zero and levels % 2 == 0:
        raise ValueError("Levels must be an odd number when centering on zero.")

    # If centering on zero, adjust the min and max to be symmetrical around zero
    if centered_on_zero:
        max_abs_val = max(abs(min_val), abs(max_val))
        min_val, max_val = -max_abs_val, max_abs_val

    # Calculate the step size
    step = (max_val - min_val) / (levels - 1)

    # Round the step to the nearest number ending in 0, 2, or 5
    step = round_step(step)

    # Adjust min and max values to fit the new step size
    if centered_on_zero:
        # Ensure that min_val is a multiple of the step size away from zero
        min_val = -((levels - 1) // 2) * step
        max_val = ((levels - 1) // 2) * step
    else:
        min_val = np.floor(min_val / step) * step
        max_val = min_val + step * (levels - 1)

    # Generate the values and boundaries
    values = np.linspace(min_val, max_val, levels)
    boundaries = [val - step / 2 for val in values] + [max_val + step / 2]

    base_cmap = plt.get_cmap(cmap_name)
    colors = base_cmap(np.linspace(0, 1, levels))

    if white_center and centered_on_zero:
        # Set the middle interval color to white
        mid_index = (levels - 1) // 2
        colors[mid_index] = (1, 1, 1, 1)

    cmap = ListedColormap(colors)
    norm = BoundaryNorm(boundaries, cmap.N, clip=True)

    return cmap, norm


def update_stats(
    var1,
    var2,
    diffvar,
    weights,
    ydim,
    xdim,
    ordinate,
    abscissa,
    singlepanel,
    plot_type,
    projection,
    splitscale,
    ax4=None,
    ax5=None,
    ax8=None,
):
    """
    Function update statistics, calculate RMS, and plot zonal/meridonal
    RMS panels

    Parameters
    ----------
    var1 : xarray.DataArray
        Input data variable (A)
    var2 : xarray.DataArray
        Input data variable (B)
    diffvar : xarray.DataArray
        Difference array (A minus B)
    weights : xarray.DataArray
        Weights array to use for area averaging
    ydim : str
        Name of y-dimension (image-relative, not geocentric)
    xdim : str
        Name of x-dimension (image-relative, not geocentric)
    ordinate : str
        Name of y-coordinate (image-relative, not geocentric)
    abscissa : str
        Name of x-coordinate (image-relative, not geocentric)
    singlepanel : bool
        Difference plot only if True, otherwise three-panel format
    plot_type : str
        Either "yx" or "yz"
    projection : None or cartopy.crs.Projection
        Map projection
    splitscale : float
        Split vertical scale (yz plots only)
    ax4 : matplotlib.axes.Axes, optional
        RHS RMS plot
    ax5 : matplotlib.axes.Axes, optional
        Bottom RMS plot
    ax8 : matplotlib.axes.Axes, optional
        Stats area axis
    """
    # Calculate RMSE
    var_se = diffvar**2
    var_rmse_xave = np.sqrt(var_se.weighted(weights).mean(xdim))
    var_rmse_yave = np.sqrt(var_se.weighted(weights).mean(ydim))
    rms = float(np.sqrt(var_se.weighted(weights).mean((ydim, xdim))))

    # Plot RMS "wings" only in three-panel mode
    if not singlepanel:
        if ax4 is not None:
            # Generate y nominal coordinates
            if len(diffvar[ordinate].shape) == 2:
                nominal_y = diffvar[ordinate].mean(xdim)
            else:
                nominal_y = diffvar[ordinate]

            # Plot the zonal (yx) or depth (yz) mean RMS difference
            if projection is None:
                ax4.plot(var_rmse_xave, nominal_y, color="k")
                ax4.set_ylim(nominal_y.min(), nominal_y.max())
                ax4.text(
                    0.5,
                    1.01,
                    "RMSE",
                    ha="center",
                    fontsize=8,
                    transform=ax4.transAxes,
                )
                if plot_type == "yz":
                    if splitscale is not None:
                        ax4.set_yscale("splitscale", zval=splitscale)
                        ax4.axhline(
                            y=splitscale[1],
                            color="black",
                            linewidth=0.5,
                            linestyle="dashed",
                        )
                    else:
                        ax4.invert_yaxis()
                ax4.set_yticks([])
            else:
                ax4.axis("off")

        # Plot the meridional (yx) or zonal (yz) mean RMS difference
        if ax5 is not None:
            # Generate x nominal coordinates
            if len(diffvar[ordinate].shape) == 2:
                nominal_x = diffvar[abscissa].mean(ydim)
            else:
                nominal_x = diffvar[abscissa]

            # Plot the meridional mean RMS difference
            if projection is None:
                ax5.plot(nominal_x, var_rmse_yave, color="k")
                ax5.set_xlim(nominal_x.min(), nominal_x.max())
                ax5.set_xticks([])
                ax5.yaxis.set_label_position("right")
                ax5.yaxis.tick_right()
            else:
                ax5.axis("off")

    # Add A & B stats boxes if in three panel mode
    if not singlepanel:
        add_stats_box(ax8, 0.0, 0.5, "Dataset A", float(var1.min()), float(var1.max()))
        add_stats_box(ax8, 0.15, 0.5, "Dataset B", float(var2.min()), float(var2.max()))

    # Add difference plot stats box
    add_stats_box(
        ax8,
        0.6,
        0.5,
        "Difference",
        float(diffvar.min()),
        float(diffvar.max()),
        float(diffvar.weighted(weights).mean((ydim, xdim))),
        rms,
        ha="left",
    )


def create_event_handler(
    ax,
    var1,
    var2,
    diffvar,
    weights,
    ydim,
    xdim,
    ordinate,
    abscissa,
    singlepanel,
    plot_type,
    projection,
    splitscale,
    ax4=None,
    ax5=None,
    ax8=None,
    event="zoom",
):
    def on_zoom(event):
        if event.inaxes == ax:  # Check if the event was on the correct axes
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()

            var1_sub = geoslice(var1, x=xlim, y=ylim)
            var2_sub = geoslice(var2, x=xlim, y=ylim)
            diffvar_sub = geoslice(diffvar, x=xlim, y=ylim)
            weights_sub = geoslice(weights, x=xlim, y=ylim)

            # Update the title or perform other actions based on the zoom
            # ax.set_title(f"Zoomed Range: x={xlim}, y={ylim}")

            # Clear ax8 and add new statistical information
            _ = ax8.clear()
            _ = ax8.axis("off")
            update_stats(
                var1_sub,
                var2_sub,
                diffvar_sub,
                weights_sub,
                ydim,
                xdim,
                ordinate,
                abscissa,
                singlepanel,
                plot_type,
                projection,
                splitscale,
                ax4=ax4,
                ax5=ax5,
                ax8=ax8,
            )

            plt.draw()

    def on_draw(event):
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()

        var1_sub = geoslice(var1, x=xlim, y=ylim)
        var2_sub = geoslice(var2, x=xlim, y=ylim)
        diffvar_sub = geoslice(diffvar, x=xlim, y=ylim)
        weights_sub = geoslice(weights, x=xlim, y=ylim)

        # Update the title or perform other actions based on the zoom
        # ax.set_title(f"Zoomed Range: x={xlim}, y={ylim}")

        # Clear ax8 and add new statistical information
        _ = ax8.clear()
        _ = ax8.axis("off")

        if projection is None:
            update_stats(
                var1_sub,
                var2_sub,
                diffvar_sub,
                weights_sub,
                ydim,
                xdim,
                ordinate,
                abscissa,
                singlepanel,
                plot_type,
                projection,
                splitscale,
                ax4=None,
                ax5=None,
                ax8=ax8,
            )

        plt.draw()

    if event == "zoom":
        return on_zoom
    if event == "draw":
        return on_draw


def compare_2d(
    var1,
    var2,
    label1=None,
    label2=None,
    xrng=None,
    yrng=None,
    levels=17,
    clim=None,
    clim_diff=None,
    cmap1="viridis",
    cmap2="RdBu_r",
    stats=True,
    plot_type=None,
    splitscale=None,
    projection=None,
    singlepanel=False,
    dpi=150,
    topright_label="",
):
    """Compare two datasets on a set of matplotlib subplots.

    The function assumes that the two input arrays are on the same
    horizontal (xy) grid.

    Parameters
    ----------
    var1 : xarray.DataArray
        First dataset for comparison.
    var2 : xarray.DataArray
        Second dataset for comparison.
    label1 : str, optional
        Label for the first dataset.
    label2 : str, optional
        Label for the second dataset.
    xrng : tuple, optional
        Range for x-axis slicing (min, max).
    yrng : tuple, optional
        Range for y-axis slicing (min, max).
    levels : int, optional
        Number of levels for color mapping.
    clim : tuple, optional
        Color limit for the first dataset's plot.
    clim_diff : tuple, optional
        Color limit for the difference plot.
    cmap1 : str, optional
        Colormap for the first dataset's plot.
    cmap2 : str, optional
        Colormap for the difference plot.
    stats : bool, optional
        Calculate bias and RMS difference, by default True
    plot_type : str, optional
        Either "yx" or "yz", value inferred if missing
    splitscale : float, optional
        Depth value for vertical split scale, by default None
    projection : cartopy.crs.Projection, optional
        Cartopy projection, by default None
    dpi : float
        Dots per inch (dpi) value to use for image
    topright_label :

    Returns
    -------
    matplotlib.figure.Figure
        The matplotlib figure object with the plots.
    """

    # Set the overall figure size, maintain a 16:9 aspect ratio, and set dpi
    fig = plt.figure(figsize=(16, 9), dpi=dpi, facecolor="white")

    # Create a grid layout.
    grid = plt.GridSpec(16, 16, figure=fig)

    # Define a facecolor for the subplots. This color will be used to fill in
    # missing values and be the deafult land color
    facecolor = "gray"

    # Define the subplot axes
    if singlepanel:
        # Used for a single-panel difference plot
        ax0 = fig.add_subplot(grid[0, :])  # Header
        ax3 = fig.add_subplot(
            grid[1:12, 1:16], facecolor=facecolor, projection=projection
        )  # Right
        ax4 = None
        ax5 = None
        ax7 = fig.add_subplot(grid[13:14, :])  # Right Colorbar
        ax8 = fig.add_subplot(grid[14:16, :])  # Info panel

    else:
        # The structure of the three-panel plot:
        #   The left side has plots of the two input variables that are stacked
        #   vertically. The right side has a difference plot that is larger
        #   in order to emphasize the difference between the two variables

        ax0 = fig.add_subplot(grid[0, :])  # Header
        ax1 = fig.add_subplot(
            grid[1:8, :6], facecolor=facecolor, projection=projection
        )  # Left Top
        ax2 = fig.add_subplot(
            grid[8:15, :6], facecolor=facecolor, projection=projection
        )  # Left Bottom
        ax3 = fig.add_subplot(
            grid[1:12, 6:15], facecolor=facecolor, projection=projection
        )  # Right
        ax4 = fig.add_subplot(grid[1:12, 15:16])  # Zonal Means
        ax5 = fig.add_subplot(grid[12:13, 6:15])  # Meridional Means
        ax6 = fig.add_subplot(grid[15:16, :6])  # Left Colorbar
        ax7 = fig.add_subplot(grid[13:14, 6:15])  # Right Colorbar
        ax8 = fig.add_subplot(grid[14:16, 6:])  # Info panel

    # Now that the projections are initialized, set the supplied cartopy
    # projection to PlateCarree to maximize plotting compatibility
    if projection is not None:
        import cartopy.crs as ccrs

        projection = ccrs.PlateCarree()

    # Turn off the bounding boxes for the header and footer space
    _ = ax0.axis("off")
    _ = ax8.axis("off")

    # Infer the names of the grid coordinates and dimensions
    if plot_type is None:
        _var = var1.reset_coords(drop=True)
        if "longitude" in _var.cf.coordinates:
            plot_type = "yx"
        elif "vertical" in _var.cf.coordinates:
            plot_type = "yz"
        else:
            raise ValueError("Unable to determine if this yx or yz plot")

    # The `_var` object is used below to infer coordinates. The `var1`
    # and `var2` objects have already been checked that they are identical
    # at this point

    _var = var1

    # The plots are 2-dimensional. Determine the coordinate variable names
    # that should be used for plotting
    if plot_type == "yx":
        abscissa = _var.cf.coordinates["longitude"][0]
        ordinate = _var.cf.coordinates["latitude"][0]

    elif plot_type == "yz":
        abscissa = _var.cf.coordinates["latitude"][0]
        ordinate = _var.cf.coordinates["vertical"][0]

    # Determine the plot-relative (not grid-relative) x and y dimension names
    ydim = _var.dims[-2]
    xdim = _var.dims[-1]

    # If a yz-plot is requested, determine the bottom level to use if setting
    # a "splitscale" level
    if plot_type == "yz":
        if splitscale is not None:
            zbot = float(min(_var[ydim].max(), 6500))
            splitscale = [zbot, splitscale, 0.0]

    # Attempt to determine the metadata from the first variable
    var_name = var1.name
    long_name = var1.long_name if "long_name" in var1.attrs.keys() else ""
    units = var1.units if "units" in var1.attrs.keys() else ""

    # Add annotations at the top of the figure for the variable metadata
    _ = ax0.text(
        0.0, 0.8, var_name, weight="bold", fontsize=16, transform=ax0.transAxes
    )
    _ = ax0.text(
        0.0,
        0.0,
        long_name,
        style="italic",
        fontsize=12,
        transform=ax0.transAxes,
    )

    # Set string labels for the input datasets, also define formatting
    # properties for the label boxes
    label1 = "" if label1 is None else f" {str(label1)}"
    label2 = "" if label2 is None else f" {str(label2)}"
    props = dict(
        boxstyle="round,pad=0.3", edgecolor="black", linewidth=1.5, facecolor="white"
    )

    # Drop portions if coordinates are NaNs. This can happen especially in yz plots
    # when selecting a domain (e.g. "atlarc")  The masking/averaging can introduce
    # NaNs in the coordinate variable
    var1 = var1.where(~var1[abscissa].isnull(), drop=True)
    var2 = var2.where(~var2[abscissa].isnull(), drop=True)

    # Subset a region of the variables if requested
    if isinstance(xrng, tuple) and isinstance(yrng, tuple):
        var1 = geoslice(var1, x=xrng, y=yrng)
        var2 = geoslice(var2, x=xrng, y=yrng)
        geosliced = True
    else:
        geosliced = False

    if not singlepanel:
        # The left-hand upper and lower panels are plotted in three-panel mode;
        # these panels are turned off if only a difference plot is requested.

        # Figure out the colorbar range based on data from both variables
        if clim is None:
            _ = np.concatenate((var1, var2))
            clim = (np.nanpercentile(_, 2), np.nanpercentile(_, 98))
        cmap, norm = generate_cmap_and_norm(*clim, levels, cmap_name=cmap1)

        # Top left panel - Dataset "A"
        if projection is not None:
            cb1 = ax1.pcolormesh(
                var1[abscissa],
                var2[ordinate],
                var1.values,
                cmap=cmap,
                norm=norm,
                transform=projection,
            )
            ax1.coastlines(linewidth=0.5)

        else:
            cb1 = ax1.pcolormesh(
                var1[abscissa], var2[ordinate], var1.values, cmap=cmap, norm=norm
            )

        ax1.text(
            0.05,
            0.05,
            f"A.{label1}",
            transform=ax1.transAxes,
            fontsize=8,
            ha="left",
            bbox=props,
        )

        # Bottom left panel - Dataset "B"
        if projection is not None:
            cb2 = ax2.pcolormesh(
                var2[abscissa],
                var2[ordinate],
                var2.values,
                cmap=cmap,
                norm=norm,
                transform=projection,
            )
            ax2.coastlines(linewidth=0.5)

        else:
            cb2 = ax2.pcolormesh(
                var2[abscissa], var2[ordinate], var2.values, cmap=cmap, norm=norm
            )

        ax2.text(
            0.05,
            0.05,
            f"B.{label2}",
            transform=ax2.transAxes,
            fontsize=8,
            ha="left",
            bbox=props,
        )

        # Add the colorbar
        fig.colorbar(cb1, cax=ax6, orientation="horizontal", extend="both", label=units)

    # Right Panel - Difference plot
    # Calculate a difference field
    # Determine range and colorbar for difference plot
    # Then plot the field and add annotations

    diffvar = var1 - var2

    if clim_diff is None:
        clim_diff = (np.nanpercentile(diffvar, 2), np.nanpercentile(diffvar, 98))

    try:
        levels_m1 = levels - 1 if levels % 2 == 0 else levels
        cmap, norm = generate_cmap_and_norm(
            *clim_diff,
            levels_m1,
            cmap_name=cmap2,
            centered_on_zero=True,
            white_center=True,
        )
    except Exception as exc:
        warnings.warn(str(exc))
        cmap = cmap2
        norm = None

    if projection is not None:
        cb3 = ax3.pcolormesh(
            diffvar[abscissa],
            diffvar[ordinate],
            diffvar.values,
            cmap=cmap,
            norm=norm,
            transform=projection,
        )
        ax3.coastlines(linewidth=0.5)

        if geosliced:
            ax3.set_extent([*xrng, *yrng], crs=projection)

    else:
        cb3 = ax3.pcolormesh(
            diffvar[abscissa],
            diffvar[ordinate],
            diffvar.values,
            cmap=cmap,
            norm=norm,
        )

    ax3.text(
        0.05,
        0.05,
        "Difference (A minus B)",
        transform=ax3.transAxes,
        fontsize=12,
        ha="left",
        bbox=props,
    )

    ax3.text(
        0.98,
        1.03,
        topright_label,
        transform=ax3.transAxes,
        fontsize=10,
        style="italic",
        ha="right",
    )

    # Add the colorbar for the difference plot
    fig.colorbar(cb3, cax=ax7, orientation="horizontal", extend="both", label=units)

    axes = [ax3] if singlepanel else [ax1, ax2, ax3]

    if plot_type == "yz":
        for ax in axes:
            if splitscale is not None:
                ax.set_yscale("splitscale", zval=splitscale)
                ax.axhline(
                    y=splitscale[1], color="black", linewidth=0.5, linestyle="dashed"
                )
            else:
                ax.invert_yaxis()

        plt.subplots_adjust(wspace=2.0, hspace=1.5)

    else:
        # Remove ticks and labels
        for ax in axes:
            ax.set_xticks([])
            ax.set_yticks([])

        plt.subplots_adjust(wspace=0.7, hspace=0.8)

    if plot_type == "yz":
        zbounds = infer_bounds(diffvar[ordinate].values, 0, None)
        dz = zbounds[1:] - zbounds[0:-1]

        ybounds = infer_bounds(diffvar[abscissa].values, -90, 90)
        ybounds = np.radians(ybounds)

        upper_y = ybounds[1:]
        lower_y = ybounds[0:-1]

        dy = (
            2.0
            * 6371.0e3
            * np.arcsin(np.sqrt(np.sin((upper_y - lower_y) / 2.0) ** 2.0))
        )

        weights = xr.DataArray(
            dz[:, np.newaxis] * dy, dims=diffvar.dims, coords=diffvar.coords
        )

    else:
        # Infer the cell area name
        area = var1.cf.standard_names["cell_area"][0]
        weights = var1[area]

    if stats:
        update_stats(
            var1,
            var2,
            diffvar,
            weights,
            ydim,
            xdim,
            ordinate,
            abscissa,
            singlepanel,
            plot_type,
            projection,
            splitscale,
            ax4=ax4,
            ax5=ax5,
            ax8=ax8,
        )

    else:
        if not singlepanel:
            ax4.axis("off")
            ax5.axis("off")

    if singlepanel and plot_type == "yx":
        draw_handler = create_event_handler(
            ax3,
            var1,
            var2,
            diffvar,
            weights,
            ydim,
            xdim,
            ordinate,
            abscissa,
            singlepanel,
            plot_type,
            projection,
            splitscale,
            ax4=ax4,
            ax5=ax4,
            ax8=ax8,
            event="draw",
        )

        if projection is not None:
            fig.canvas.mpl_connect("button_release_event", draw_handler)

        else:
            fig.canvas.mpl_connect("draw_event", draw_handler)

    return fig



"""classes.py : momgrid object definitions """

__all__ = ["MOMgrid", "Gridset"]

from momgrid.external import woa18_grid

from momgrid.metadata import add_metadata

from momgrid.util import (
    associate_grid_with_data,
    get_file_type,
    is_hgrid,
    is_static,
    is_symmetric,
    load_cached_grid,
    read_netcdf_from_tar,
    reset_nominal_coords,
    standard_grid_area,
)

from momgrid.external import build_regridder_weights, static_to_xesmf

import xarray as xr
import numpy as np
import os
import pickle
import warnings

KNOWN_GRIDS = [
    "om4_sym",
    "esm4_sym",
    "om4_nonsym",
    "esm4_nonsym",
    "om5",
    "nwa12",
    "cm4x",
    "spearmed_nonsym",
]


class MOMgrid:
    def __init__(
        self,
        source,
        topog=None,
        depth_var="depth",
        symmetric=True,
        verbose=False,
        hgrid_dtype="float32",
        max_depth=6500.0,
        warn=True,
    ):
        """Ocean Model grid object

        Parameters
        ----------
        source : xarray.Dataset or str, path-like
            Xarray dataset or path to grid source, either an ocean_hgrid.nc
            file or an ocean_static.nc file. Optionally, a known configuration
            may also be specified.
        topog : xarray.Dataset or str, path-like, optional
            Xarray dataset or path to ocean topography. If present the wet masks
            corresponding to the grids will be generated, by default None.
        depth_var : str, optional
            Name of the depth variable in the topog file, by default "depth"
        symmetric : bool, optional
            Return metrics compatible with symmetric memory mode,
            by default True
        verbose : bool, optional
            Verbose output, by default False
        hgrid_dtype : str, numpy.dtype
            Precision to use when reconstructing fields from the
            `ocean_hgrid.nc`. The supergrid is natively double precision
            while the ocean static files are single precision,
            by default "float32"
        max_depth : float, optional
            by default 6500.
        warn : bool, optional
            Issue warnings, by default True
        """

        # Define names in static file for future flexibility
        geolon = "geolon"
        geolat = "geolat"
        areacello = "areacello"
        dxt = "dxt"
        dyt = "dyt"
        deptho = "deptho"
        wet = "wet"

        # Define names in supergrid file for future flexibility
        xvar = "x"
        yvar = "y"
        areavar = "area"
        dxvar = "dx"
        dyvar = "dy"

        # Specified as kwarg
        depth = depth_var

        # Load source file
        if isinstance(source, xr.Dataset):
            ds = source

        elif isinstance(source, str):
            if source in KNOWN_GRIDS:
                ds = load_cached_grid(source).to_xarray()

            else:
                abspath = os.path.abspath(source)
                if os.path.exists(abspath):
                    self.source = abspath
                    ftype = get_file_type(abspath)
                    if ftype == "netcdf":
                        ds = xr.open_dataset(source)
                    elif ftype == "tar":
                        ds = read_netcdf_from_tar(abspath, "ocean_hgrid.nc")

                        try:
                            topog = read_netcdf_from_tar(abspath, "ocean_topog.nc")
                        except Exception as _:
                            try:
                                topog = read_netcdf_from_tar(abspath, "topog.nc")
                            except Exception as _:
                                warning.warn(
                                    "Unable to find topog in gridspec bundle. "
                                    + "Not processing wet masks or deptho"
                                )

                    else:
                        raise ValueError(f"Unknown input file type for {abspath}")

                else:
                    raise ValueError(f"Unknown source: {source}")

        else:
            raise ValueError(
                "Source must be an xarray dataset, path, or known model config."
            )

        # Load topog file
        if (isinstance(topog, xr.Dataset)) or (topog is None):
            self.topog_source = str(type(topog))
            topog = topog

        elif isinstance(topog, str):
            abspath = os.path.abspath(topog)
            if os.path.exists(abspath):
                self.topog_source = abspath
                topog = xr.open_dataset(topog)
            else:
                raise ValueError(f"Unknown source: {source}")

        else:
            raise ValueError(
                "Source must be an xarray dataset, path, or known model config."
            )

        # TODO: remove this object below; only used in testing
        self.ds = ds

        # Store whether or not the source is a static file or an hgrid file
        self.is_hgrid = is_hgrid(ds)
        self.is_static = is_static(ds)

        # If a static file is defined, test if it is compatible with the symmetric
        # memory mode that was requested
        if self.is_static:
            if is_symmetric(ds) is not symmetric:
                if warn:
                    warnings.warn(
                        f"Supplied static file inconsistent with requested memory mode. Adjusting ..."
                    )
                self.symmetric = is_symmetric(ds)
            else:
                self.symmetric = symmetric
        else:
            self.symmetric = symmetric

        # If the source information is from the hgrid file, pre-load data
        if self.is_hgrid:
            # TODO: add a hook here to do the downsample

            # x and y values and distances
            x = ds[xvar].values
            dx = ds[dxvar].values
            y = ds[yvar].values
            dy = ds[dyvar].values

            # make cell areas periodic in x and y
            area = ds["area"].values
            area = np.append(area[0, :][None, :], area, axis=0)
            area = np.append(area[:, 0][:, None], area, axis=1)

        # Fetch tracer cell grid metrics
        if self.is_static:
            for x in [geolon, geolat, areacello, deptho, wet, dxt, dyt]:
                try:
                    setattr(self, x, ds[x].values)
                except Exception as exc:
                    if warn:
                        warnings.warn(f"Unable to load {x}")

        elif self.is_hgrid:
            setattr(self, geolon, x[1::2, 1::2].astype(hgrid_dtype))
            setattr(self, dxt, (dx[1::2, ::2] + dx[1::2, 1::2]).astype(hgrid_dtype))

            setattr(self, geolat, y[1::2, 1::2].astype(hgrid_dtype))
            setattr(self, dyt, (dy[::2, 1::2] + dy[1::2, 1::2]).astype(hgrid_dtype))

            _area = area[:-1, :-1]
            _area = (
                _area[::2, ::2]
                + _area[1::2, 1::2]
                + _area[::2, 1::2]
                + _area[1::2, ::2]
            )
            setattr(self, areacello, _area.astype(hgrid_dtype))

            if topog is not None:
                _depth = topog[depth].values
                _depth = np.where(_depth > max_depth, max_depth, _depth)
                _depth = np.where(_depth > 0, _depth, np.nan)
                _wet = np.where(np.isnan(_depth), 0.0, 1.0)

                setattr(self, deptho, _depth.astype(hgrid_dtype))
                setattr(self, wet, _wet.astype(hgrid_dtype))

                # reflect top row about the center
                _wet_padded = np.concatenate((_wet, _wet[-1, :][::-1][None, :]), axis=0)

                _wet_padded = np.concatenate(
                    (_wet_padded, _wet_padded[:, 0][:, None]), axis=1
                )

        # Fetch u-cell grid metrics
        suffix = "_u"
        if self.is_static:
            for x in [
                geolon + suffix,
                geolat + suffix,
                areacello + suffix,
                "dxCu",
                "dyCu",
                wet + suffix,
            ]:
                try:
                    setattr(self, x, ds[x].values)
                except Exception as exc:
                    if warn:
                        warnings.warn(f"Unable to load {x}")

        elif self.is_hgrid:
            _geolon = x[1::2, ::2]
            _geolon = _geolon if self.symmetric else _geolon[:, 1:]
            setattr(self, geolon + suffix, _geolon.astype(hgrid_dtype))

            _dxCu = dx[1::2, ::2]
            _dxCu = _dxCu + np.roll(dx[1::2, :-1:2], -1, axis=-1)
            _dxCu = np.append(_dxCu[:, 0][:, None], _dxCu, axis=1)
            _dxCu = _dxCu if self.symmetric else _dxCu[:, 1:]
            setattr(self, "dxCu", _dxCu.astype(hgrid_dtype))

            _geolat = y[1::2, ::2]
            _geolat = _geolat if self.symmetric else _geolat[:, 1:]
            setattr(self, geolat + suffix, _geolat.astype(hgrid_dtype))

            _dyCu = dy[::2, 2::2] + dy[1::2, 2::2]
            _dyCu = np.append(_dyCu[:, 0][:, None], _dyCu, axis=1)
            _dyCu = _dyCu if self.symmetric else _dyCu[:, 1:]
            setattr(self, "dyCu", _dyCu.astype(hgrid_dtype))

            _area = area[:-1, :]
            _area = (
                _area[::2, 1::2]
                + _area[1::2, 2::2]
                + _area[::2, 2::2]
                + _area[1::2, 1::2]
            )
            _area = np.append(_area[:, 0][:, None], _area, axis=1)
            _area = _area if self.symmetric else _area[:, 1:]
            setattr(self, areacello + suffix, _area.astype(hgrid_dtype))

            if topog is not None:
                _wet = np.minimum(np.roll(_wet_padded, 1, axis=1), _wet_padded)
                _wet = _wet if not self.symmetric else _wet[0:-1, :]
                setattr(self, wet + suffix, _wet.astype(hgrid_dtype))

        # Fetch v-cell grid metrics
        suffix = "_v"
        if self.is_static:
            for x in [
                geolon + suffix,
                geolat + suffix,
                areacello + suffix,
                "dxCv",
                "dyCv",
                wet + suffix,
            ]:
                try:
                    setattr(self, x, ds[x].values)
                except Exception as exc:
                    if warn:
                        warnings.warn(f"Unable to load {x}")

        elif self.is_hgrid:
            _geolon = x[::2, 1::2]
            _geolon = _geolon if self.symmetric else _geolon[1:, :]
            setattr(self, geolon + suffix, _geolon.astype(hgrid_dtype))

            _dxCv = dx[2::2, ::2] + dx[2::2, 1::2]
            _dxCv = np.append(_dxCv[0, :][None, :], _dxCv, axis=0)
            _dxCv = _dxCv if self.symmetric else _dxCv[1:, :]
            setattr(self, "dxCv", _dxCv.astype(hgrid_dtype))

            _geolat = y[::2, 1::2]
            _geolat = _geolat if self.symmetric else _geolat[1:, :]
            setattr(self, geolat + suffix, _geolat.astype(hgrid_dtype))

            _dyCv = dy[::2, 2::2] + dy[1::2, 2::2]
            _dyCv = np.append(_dyCv[0, :][None, :], _dyCv, axis=0)
            _dyCv = _dyCv if self.symmetric else _dyCv[1:, :]
            setattr(self, "dyCv", _dyCv.astype(hgrid_dtype))

            _area = area[:, :-1]
            _area = (
                _area[1::2, ::2]
                + _area[2::2, 1::2]
                + _area[1::2, 1::2]
                + _area[2::2, ::2]
            )
            _area = np.append(_area[0, :][None, :], _area, axis=0)
            _area = _area if self.symmetric else _area[1:, :]
            setattr(self, areacello + suffix, _area.astype(hgrid_dtype))

            if topog is not None:
                _wet = np.minimum(np.roll(_wet_padded, 1, axis=0), _wet_padded)
                _wet = _wet if not self.symmetric else _wet[:, 0:-1]
                setattr(self, wet + suffix, _wet.astype(hgrid_dtype))

        # Fetch corner cell grid metrics
        suffix = "_c"
        if self.is_static:
            # TODO: setattr(self, wet + suffix, ds[wet + suffix].values)
            # note: dx and dy are not defined in ocean_static.nc files
            for x in [geolon + suffix, geolat + suffix, areacello + suffix]:
                try:
                    setattr(self, x, ds[x].values)
                except Exception as exc:
                    if warn:
                        warnings.warn(f"Unable to load {x}")

        elif self.is_hgrid:
            _geolon = x[::2, ::2]
            _geolon = _geolon if self.symmetric else _geolon[1:, 1:]
            setattr(self, geolon + suffix, _geolon.astype(hgrid_dtype))

            _geolat = y[::2, ::2]
            _geolat = _geolat if self.symmetric else _geolat[1:, 1:]
            setattr(self, geolat + suffix, _geolat.astype(hgrid_dtype))

            _area = area
            _area = (
                _area[1::2, 1::2]
                + _area[2::2, 2::2]
                + _area[1::2, 2::2]
                + _area[2::2, 1::2]
            )
            _area = np.append(_area[0, :][None, :], _area, axis=0)
            _area = np.append(_area[:, 0][:, None], _area, axis=1)
            _area = _area if self.symmetric else _area[1:, 1:]
            setattr(self, areacello + suffix, _area.astype(hgrid_dtype))

            # TODO: add wet mask for corner cells

    def to_xarray(self):
        # Define dimension names for future flexibility
        ycenter = "yh"
        xcenter = "xh"
        ycorner = "yq"
        xcorner = "xq"

        # Define names in static file for future flexibility
        geolon = "geolon"
        geolat = "geolat"
        areacello = "areacello"
        dxt = "dxt"
        dyt = "dyt"
        deptho = "deptho"
        wet = "wet"

        tcell = ("", (ycenter, xcenter))
        ucell = ("_u", (ycenter, xcorner))
        vcell = ("_v", (ycorner, xcenter))
        ccell = ("_c", (ycorner, xcorner))

        cell_types = [tcell, ucell, vcell, ccell]

        ds = xr.Dataset()

        for cell_type in cell_types:
            ds[geolon + cell_type[0]] = xr.DataArray(
                getattr(self, geolon + cell_type[0]), dims=cell_type[1]
            )
            ds[geolat + cell_type[0]] = xr.DataArray(
                getattr(self, geolat + cell_type[0]), dims=cell_type[1]
            )
            ds[areacello + cell_type[0]] = xr.DataArray(
                getattr(self, areacello + cell_type[0]), dims=cell_type[1]
            )

            # TODO: The dx/dy names do not follow a predicitable pattern. Need to deal with this
            # ds[dxt+cell_type[0]] = xr.DataArray(getattr(self,dxt+cell_type[0]), dims=cell_type[1])
            # ds[dyt+cell_type[0]] = xr.DataArray(getattr(self,dyt+cell_type[0]), dims=cell_type[1])

        # TODO: Add variable attributes -- long_name, standard_name, units, etc.

        if hasattr(self, "deptho"):
            if self.deptho is not None:
                ds[deptho] = xr.DataArray(getattr(self, deptho), dims=tcell[1])

        for cell_type in cell_types:
            try:
                if hasattr(self, wet + cell_type[0]):
                    ds[wet + cell_type[0]] = xr.DataArray(
                        getattr(self, wet + cell_type[0]), dims=cell_type[1]
                    )
            except:
                # warnings.warn(f"Unable to add wet_{cell_type[0]}")
                pass

        # Promote dimensions to coords
        for coord in ["xh", "yh", "xq", "yq"]:
            ds[coord] = xr.DataArray(ds[coord], dims=(coord), coords={coord: ds[coord]})
        ds = reset_nominal_coords(ds)

        # Add variable metadata
        ds = add_metadata(ds)

        return ds

    def to_xesmf(self, grid_type="t", filename=None):
        return static_to_xesmf(self.to_xarray(), grid_type=grid_type, filename=filename)

    def associate(self, data):
        return associate_grid_with_data(self.to_xarray(), reset_nominal_coords(data))

    def generate_weights(self, dsout, grid_type=["t", "u", "v", "c"], periodic=None):
        grid_type = list(grid_type) if not isinstance(grid_type, list) else grid_type
        symmetric = "sym" if self.symmetric else "nosym"

        if "t" in grid_type:
            _periodic = True if periodic is None else periodic
            dsin = self.to_xesmf(grid_type="t")
            files = build_regridder_weights(dsin, dsout, periodic=_periodic)
            _ = [os.rename(x, f"t_{symmetric}_{x}") for x in files]

        if "u" in grid_type:
            _periodic = False if periodic is None else periodic
            dsin = self.to_xesmf(grid_type="u")
            files = build_regridder_weights(dsin, dsout, periodic=_periodic)
            _ = [os.rename(x, f"u_{symmetric}_{x}") for x in files]

        if "v" in grid_type:
            _periodic = True if periodic is None else periodic
            dsin = self.to_xesmf(grid_type="v")
            files = build_regridder_weights(dsin, dsout, periodic=_periodic)
            _ = [os.rename(x, f"v_{symmetric}_{x}") for x in files]

        if "c" in grid_type:
            _periodic = False if periodic is None else periodic
            dsin = self.to_xesmf(grid_type="c")
            files = build_regridder_weights(dsin, dsout, periodic=_periodic)
            _ = [os.rename(x, f"c_{symmetric}_{x}") for x in files]

        return "Finished generating weights."


class Gridset:
    def __init__(self, dset, ignore=None):
        """Combination class of MOM grid object and a model data set

        Parameters
        ----------
        dset : xarray.Dataset, str (path-like), or list (paths)
            Model dataset object, path to NetCDF file, or list of files
        ignore : str or list
            List of variables to ignore when processing the dataset
        """

        # Open dataset if an xarray Dataset object is not supplied
        if not isinstance(dset, xr.Dataset):
            dset = xr.open_mfdataset(dset, use_cftime=True)

        # Filter variables supplied by the ignore kwarg
        if ignore is not None:
            ignore = [ignore] if not isinstance(ignore, list) else ignore
            ignore = [x for x in ignore if x in dset.keys()]
            dset = dset.drop_vars(ignore)

        # Get dimension lengths. Note the MOM6 conventions of
        # xh, yh, xq, and yq are used below
        self.xh = len(dset["xh"]) if "xh" in dset.dims else 0
        self.yh = len(dset["yh"]) if "yh" in dset.dims else 0
        self.xq = len(dset["xq"]) if "xq" in dset.dims else 0
        self.yq = len(dset["yq"]) if "yq" in dset.dims else 0

        # Construct a list of tuples of the grids inferred from the
        # supplied input dataset
        self.grid = [
            (self.yh, self.xh),
            (self.yh, self.xq),
            (self.yq, self.xh),
            (self.yq, self.xq),
        ]

        # Infer the model type based on the grid dimensions

        if any(x in [(1080, 1441), (1081, 1440), (1081, 1441)] for x in self.grid):
            self.model = "om4_sym"
            self._periodic = True

        elif any(x in [(576, 721), (577, 720), (577, 721)] for x in self.grid):
            self.model = "esm4_sym"
            self._periodic = True

        # Note that if a dataset contains ONLY variables in the tracer grid,
        # it would be better to use a symmetric version of the grid since it
        # correctly includes the corner cell data
        elif any(x in [(1080, 1440)] for x in self.grid):
            self.model = "om4_nonsym"
            self._periodic = True

        # Note that if a dataset contains ONLY variables in the tracer grid,
        # it would be better to use a symmetric version of the grid since it
        # correctly includes the corner cell data
        elif any(x in [(576, 720)] for x in self.grid):
            self.model = "esm4_nonsym"
            self._periodic = True

        elif any(
            x in [(1161, 1440), (1161, 1441), (1162, 1440), (1162, 1441)]
            for x in self.grid
        ):
            self.model = "om5"
            self._periodic = True

        elif any(
            x in [(845, 775), (845, 776), (846, 775), (846, 776)] for x in self.grid
        ):
            self.model = "nwa12"
            self._periodic = False

        elif any(
            x in [(2240, 2880), (2240, 2881), (2241, 2880), (2241, 2881)]
            for x in self.grid
        ):
            self.model = "cm4x"
            self._periodic = True

        elif any(x in [(320, 360)] for x in self.grid):
            self.model = "spearmed_nonsym"
            self._periodic = True

        else:
            self.model = None
            self._periodic = None

        # Load a cached copy of the grid object for the specific model.
        # This cached copy is significantly faster than calculating each time.
        # However, it might be a good idea to allow run-time calculation
        # as an option in the future for new or unsupported model configs.

        if self.model is not None:
            self.grid = load_cached_grid(self.model)
        else:
            raise ValueError("Unable to infer model from input data.")

        # Associate the 2D geolat/geolon as coordinate variables
        self.data = self.grid.associate(dset)

        # Drop singleton dimensions and coordinates
        self.data = self.data.squeeze()

        # Re-apply metadata for coordinate variables
        self.coord_attrs = {}
        for coord in self.data.coords:
            self.coord_attrs[coord] = self.data[coord].attrs

    def shape_string(self, grid_type):
        """Constructs xESMF string for cached grid weights"""
        if grid_type == "t":
            shape = f"{self.yh}x{self.xh}"
        elif grid_type == "u":
            shape = f"{self.yh}x{self.xq}"
        elif grid_type == "v":
            shape = f"{self.yq}x{self.xh}"
        elif grid_type == "c":
            shape = f"{self.yq}x{self.xq}"
        else:
            shape = ""

        return shape

    def grid_type(self, dims):
        """Infers grid type based on dimension names"""
        if set(("yh", "xh")).issubset(dims):
            grid_type = "t"
        elif set(("yh", "xq")).issubset(dims):
            grid_type = "u"
        elif set(("yq", "xh")).issubset(dims):
            grid_type = "v"
        elif set(("yq", "xq")).issubset(dims):
            grid_type = "c"
        else:
            grid_type = ""

        return grid_type

    def periodic(self, grid_type):
        """Returns False for u and c grids"""
        return False if grid_type in ["u", "c"] else self._periodic

    def regrid_var(self, var, method="bilinear", resolution=1.0):
        """Regrids a variable using xESMF

        Parameters
        ----------
        var : str
            Variable name
        method : str, optional
            xESMF regridding method, by default "bilinear"
        resolution : float
            Target WOA18 grid resolution (1.0 or 0.25), by default 0.25

        Returns
        -------
        xarray.DataArray
        """

        # Import deliberately added here. (Did not want to make the package
        # dependent on xESMF -- just needed for regridding)

        import xesmf as xe

        # Determine grid type
        grid_type = self.grid_type(self.data[var].dims)
        symmetric = "sym" if self.grid.symmetric else "nosym"

        # Determine if grid is periodic
        if method not in ["conservative"]:
            periodic = "_peri" if self.periodic(grid_type) else ""
        else:
            periodic = ""

        # Constuct xESMF shape string
        shape = self.shape_string(grid_type)

        # Setup output grid

        if resolution == 1.0:
            grid_dst = woa18_grid(resolution=1.0)
            dims_dst = "180x360"

        elif resolution == 0.25:
            grid_dst = woa18_grid(resolution=0.25)
            dims_dst = "720x1440"

        # Construct full name of the saved regridder weights
        regridder = f"{grid_type}_{symmetric}_{method}_{shape}_{dims_dst}{periodic}"

        if "MOMGRID_WEIGHTS_DIR" in os.environ.keys():
            weights_dir = os.environ["MOMGRID_WEIGHTS_DIR"]
        else:
            weights_dir = "./grid_weights"

        # Check if a cached version of the regridder exists. If not, load it
        if not hasattr(self, regridder):
            setattr(
                self,
                regridder,
                xe.Regridder(
                    self.grid.to_xesmf(grid_type=grid_type),
                    grid_dst,
                    method,
                    weights=f"{weights_dir}/{regridder}.nc",
                ),
            )

        # Regrid the variable using the regridder
        _regridder = getattr(self, regridder)
        result = _regridder(self.data[var])

        # Set variable name, attributes, and coordinates
        result.name = var
        result.attrs = {**result.attrs, **self.data[var].attrs}
        for coord in result.coords:
            if coord in self.data[var].coords:
                if len(result.coords[coord].attrs) == 0:
                    result[coord].attrs = self.coord_attrs[coord]

        return result

    def regrid(self, method="bilinear", resolution=1.0):
        """Loop over all variables and regrid"""

        # Create an empty dataset to hold the results
        dsout = xr.Dataset()

        # Loop over variables
        for var in list(self.data.keys()):
            try:
                dsout[var] = self.regrid_var(var, method=method, resolution=resolution)

            # Warn if regridding failed
            except Exception as exc:
                warnings.warn(str(exc))
                pass

        # Add back in the mask and the bounds
        woa_grid = woa18_grid(resolution=resolution)
        for var in ["lat_b", "lon_b", "mask"]:
            dsout[var] = woa_grid[var]
            dsout = dsout.set_coords(var)

        # Calculate the cell area and set attributes
        dsout["areacello"] = xr.DataArray(
            standard_grid_area(dsout["lat_b"], dsout["lon_b"]),
            dims=("lat", "lon"),
        )

        dsout["areacello"].attrs = {
            "units": "m2",
            "long_name": "Ocean Grid-Cell Area",
            "standard_name": "cell_area",
        }

        # Make cell area a coordinate variable
        dsout = dsout.set_coords("areacello")

        if "z_l" in dsout.coords:
            if "z_i" not in dsout.coords and "z_i" in self.data.keys():
                dsout = dsout.assign_coords({"z_i": self.data["z_i"]})

        return dsout

    def subset(self, varlist=None):
        """Function to subset the data"""

        # Make a copy of the dataset object
        dset = self.data

        if varlist is not None:
            varlist = [varlist] if isinstance(varlist, str) else varlist
            assert isinstance(varlist, list), "varlist must be a str or list"
            dset = [dset[x] for x in varlist if x in dset.keys()]
            dset = xr.Dataset({x.name: x for x in dset})

            if "z_i" not in dset.coords and "z_i" in self.data.keys():
                dset = dset.assign_coords({"z_i": self.data["z_i"]})

        # Reassign the result to the object
        self.data = dset

        # Return a list of the retained variables
        return list(self.data.keys())

    def __repr__(self):
        # TODO: write useful information about the object
        return str(self.data)

    def __str__(self):
        # TODO: write useful information about the object
        return self.data



"""util.py : auxillary functions for inferring dataset characteristics"""

__all__ = [
    "associate_grid_with_data",
    "extract_level",
    "get_file_type",
    "infer_bounds",
    "is_hgrid",
    "is_static",
    "is_symmetric",
    "load_cached_grid",
    "read_netcdf_from_tar",
    "reset_nominal_coords",
    "standard_grid_area",
    "x_average_dataset",
    "verify_dim_lens",
]

import os
import pickle
import warnings
import cmip_basins

import os.path
import tarfile
import numpy as np
import xarray as xr
from io import BytesIO

from momgrid.metadata import nominal_coord_metadata
from momgrid.geoslice import x_slice


def associate_grid_with_data(grid, data):
    """Function to associate grid metrics with data

    This function accepts a grid object and an xarray data object
    and adds the associated geolon/geolat data to each variable.

    Parameters
    ----------
    grid : xarray.Dataset
        MOMgrid-generated Xarray dataset (using the .to_xarray method)
    data : xarray.Dataset or xarray.DataArray
        MOM6 output data

    Returns
    -------
    xarray.Dataset or xarray.DataArray
    """

    # Define grid point types:
    h_point = ("yh", "xh")
    u_point = ("yh", "xq")
    v_point = ("yq", "xh")
    c_point = ("xq", "yq")

    # variables broken out in case they need to be updated later
    geolon = "geolon"
    geolat = "geolat"
    geolon_u = "geolon_u"
    geolat_u = "geolat_u"
    geolon_v = "geolon_v"
    geolat_v = "geolat_v"
    geolon_c = "geolon_c"
    geolat_c = "geolat_c"

    areacello = "areacello"
    areacello_u = "areacello_u"
    areacello_v = "areacello_v"
    areacello_c = "areacello_c"

    ds = data if isinstance(data, xr.Dataset) else xr.Dataset({data.name: data})

    # Check that dimensions are identical
    exceptions = []
    for dim in ["xh", "yh", "xq", "yq"]:
        if dim in grid.variables and dim in ds.variables:
            try:
                assert np.array_equal(grid[dim].values, ds[dim].values), dim
            except AssertionError as exc:
                exceptions.append(dim)

    if len(exceptions) > 0:
        raise RuntimeError(
            f"Cannot associate grid to data. Different dims: {exceptions}"
        )

    processed = {}

    for var in ds.keys():
        if set(h_point).issubset(ds[var].dims):
            if verify_dim_lens(ds[var], grid[geolon]):
                processed[var] = ds[var].assign_coords(
                    {
                        geolon: grid[geolon],
                        geolat: grid[geolat],
                        areacello: grid[areacello],
                    }
                )

        elif set(u_point).issubset(ds[var].dims):
            if verify_dim_lens(ds[var], grid[geolon_u]):
                processed[var] = ds[var].assign_coords(
                    {
                        geolon_u: grid[geolon_u],
                        geolat_u: grid[geolat_u],
                        areacello_u: grid[areacello_u],
                    }
                )

        elif set(v_point).issubset(ds[var].dims):
            if verify_dim_lens(ds[var], grid[geolon_v]):
                processed[var] = ds[var].assign_coords(
                    {
                        geolon_v: grid[geolon_v],
                        geolat_v: grid[geolat_v],
                        areacello_v: grid[areacello_v],
                    }
                )

        elif set(c_point).issubset(ds[var].dims):
            if verify_dim_lens(ds[var], grid[geolon_c]):
                processed[var] = ds[var].assign_coords(
                    {
                        geolon_c: grid[geolon_c],
                        geolat_c: grid[geolat_c],
                        areacello_c: grid[areacello_c],
                    }
                )

        else:
            processed[var] = ds[var]

    res = [xr.Dataset({k: v}) for k, v in processed.items()]
    res = xr.merge(res, compat="override")
    res.attrs = ds.attrs

    if isinstance(data, xr.DataArray):
        res = res[data.name]

    return res


def extract_level(dset, level):
    """Function to extract a depth level from a dataset

    This function is a wrapper for xarray's intrinsic .sel() method
    but uses cfxarray to infer the name of the vertical dimension.

    Parameters
    ----------
    dset : xarray.Dataset
        Input dataset
    level : float
        Depth level to select (m)

    Returns
    -------
    xarray.Dataset
    """

    zaxes = dset.cf.axes["Z"]

    for zax in zaxes:
        dset = dset.sel({zax: level}, method="nearest")

    return dset


def get_file_type(fname):
    """Opens a file and determines the file type based on the magic number

    The magic number for NetCDF files is 'CDF\x01' or 'CDF\x02'.
    The magic number for tar files depends on the variant but generally,
    a USTAR tar file starts with "ustar" at byte offset 257 for 5 bytes.

    Parameters
    ----------
    fname : str, path-like
        Input file string
    """

    # make sure file exists
    abspath = os.path.abspath(fname)
    assert os.path.exists(abspath), f"File does not exist: {abspath}"

    # open the file and read the first 512 bytes
    with open(abspath, "rb") as f:
        header = f.read(512)

    # look for the NetCDF magic number
    if (header[0:3] == b"CDF") or (header[1:4] == b"HDF"):
        result = "netcdf"

    # look for the tar file signature
    elif b"ustar" in header[257:262]:
        result = "tar"

    # look for gzipped file
    elif header[0:2] == b"\x1f\x8b":
        result = "tar"

    else:
        result = "unknown"

    return result


def infer_bounds(centers, start=None, end=None):
    """Function to infer cell bounds from cell centers

    This function takes a vector of cell centers. Assuming a standard grid,
    the cell bounds are inferred. Optional caps at the start and end can be
    applied if specified.

    Paramters
    ---------
    centers : np.ndarray
        Vector of cell centers
    start : float, optional
        Starting limit of bounds (e.g. -90.), by default None
    end : float, optional
        Starting limit of bounds (e.g. -90.), by default None

    Returns
    -------
    numpy.ndarray
        Vector of cell bounds with a shape of len(centers)+1
    """

    midpoints = (centers[1:] + centers[:-1]) / 2.0
    front = centers[0] - np.abs(centers[0] - midpoints[0])
    end = centers[-1] + np.abs(centers[-1] - midpoints[-1])
    midpoints = np.insert(midpoints, 0, front)
    midpoints = np.append(midpoints, end)
    return np.clip(midpoints, start, end)


def is_hgrid(ds):
    """Tests if dataset is an ocean_hgrid.nc file

    Parameters
    ----------
    ds : xarray.core.dataset.Dataset

    Returns
    -------
    bool
        True, if dataset corresponds to an hgrid file, otherwise False
    """

    # an ocean_hgrid.nc file should contain x, y, dx, and dy
    expected = set(["x", "y", "dx", "dy"])

    return expected.issubset(set(ds.variables))


def is_static(ds):
    """Tests if dataset is an ocean_static.nc file

    Parameters
    ----------
    ds : xarray.core.dataset.Dataset

    Returns
    -------
    bool
        True, if dataset corresponds to an ocean static file, otherwise False
    """

    # an ocean_static.nc file should contain at least geolon and geolat
    expected = set(["geolon", "geolat"])

    return expected.issubset(set(ds.variables))


def is_symmetric(ds, xh="xh", yh="yh", xq="xq", yq="yq"):
    """Tests if an dataset is defined on a symmetric grid

    A dataset generated in symmetric memory mode will have dimensionalty
    of `i+1` and `j+1` for the corner points compared to the tracer
    points.

    Parameters
    ----------
    ds : xarray.core.dataset.Dataset
        Input xarray dataset
    xh : str, optional
        Name of x-dimension of tracer points, by default "xh"
    yh : str, optional
        Name of y-dimension of tracer points, by default "yh"
    xq : str, optional
        Name of x-dimension of corner points, by default "xq"
    yq : str, optional
        Name of y-dimension of corner points, by default "yq"

    Returns
    -------
    bool
        True, if dataset has symmetric dimensionality, otherwise False

    """

    if set(["xh", "yh", "xq", "yq"]).issubset(ds.variables):
        xdiff = len(ds[xq]) - len(ds[xh])
        ydiff = len(ds[yq]) - len(ds[yh])

        # Basic validation checks
        assert (
            xdiff == ydiff
        ), "Diffence of tracer and corner points must be identical for x and y dimensions"
        assert xdiff in [0, 1], "Dataset is neither symmetric or non-symmetric"

        result = True if xdiff == 1 else False

    else:
        warnings.warn("Unable to determine if grid is symmetric - assuming False")
        result = False

    return result


def load_cached_grid(gridname):
    """Load a cached MOMgrid instance by name

    Parameters
    ----------
    gridname : str
        Name of grid

    Returns
    -------
    momgrid.MOMgrid
        Grid instance
    """

    if "MOMGRID_WEIGHTS_DIR" in os.environ.keys():
        weights_dir = os.environ["MOMGRID_WEIGHTS_DIR"]
    else:
        weights_dir = "./grid_weights"

    try:
        file = open(f"{weights_dir}/{gridname}.pkl", "rb")
        result = pickle.load(file)
        file.close()
    except Exception as exc:
        raise ValueError(f"Unable to load grid: {gridname}")

    return result


def read_netcdf_from_tar(tar_path, netcdf_name):
    """Reads a netcdf file from within a tar file and returns an xarray Dataset

    Parameters
    ----------
    tar_path : str, path-like
        Path to tar file
    netcdf_name : str
        Name of NetCDF file contained within the tar file

    Returns
    -------
        xarray.Dataset
            Dataset object
    """

    with open(tar_path, "rb") as f:
        tar_data = BytesIO(f.read())

    with tarfile.open(fileobj=tar_data, mode="r:*") as tar:
        if (
            netcdf_name not in tar.getnames()
            and f"./{netcdf_name}" not in tar.getnames()
        ):
            raise FileNotFoundError(
                f"The NetCDF file {netcdf_name} was not found in the tar archive."
            )

        effective_name = (
            netcdf_name if netcdf_name in tar.getnames() else f"./{netcdf_name}"
        )

        with tar.extractfile(effective_name) as netcdf_file:
            return xr.open_dataset(BytesIO(netcdf_file.read()))


def reset_nominal_coords(xobj, tracer_dims=("xh", "yh"), velocity_dims=("xq", "yq")):
    """Resets the nominal coordinate values to a monontonic series

    Tracer points are definied on the half integers while the velocity points
    are defined on the full integer points.

    Parameters
    ----------
    xobj : xarray.core.DataArray or xarray.core.Dataset
        Input xarray object
    tracer_dims : tuple, iterable, optional
        Name of tracer dimensions, by default ("xh", "yh")
    velocity_dims : tuple, iterable, optional
        Name of velocity dimensions, by default ("xq", "yq")

    Returns
    -------
        xarray.core.DataArray or xarray.core.Dataset
            Object with reset nominal coordinates
    """

    _xobj = xobj.copy()
    for dim in tracer_dims:
        if dim in _xobj.coords:
            _xobj = _xobj.assign_coords(
                {dim: list(np.arange(0.5, len(_xobj[dim]) + 0.5, 1.0))}
            )

    for dim in velocity_dims:
        if dim in _xobj.coords:
            _xobj = _xobj.assign_coords(
                {dim: list(np.arange(1.0, len(_xobj[dim]) + 1.0, 1.0))}
            )

    _xobj = nominal_coord_metadata(_xobj)

    return _xobj


def standard_grid_area(lat_b, lon_b, rad_earth=6371.0e3):
    """Function to calculate the cell areas for a standard grid

    Parameters
    ----------
    lat_b : list or numpy.ndarray
        1-D vector of latitude cell bounds
    lon_b : list or numpy.ndarray
        1-D vector of longitude cell bounds
    rad_earth : float, optional
        Radius of the Earth in meters, by default 6371.0e3

    Returns
    -------
    numpy.ndarray
        2-dimensional array of cell areas
    """

    lat_b = np.array(lat_b)
    lon_b = np.array(lon_b)

    sin_lat_b = np.sin(np.radians(lat_b))

    dy = np.abs(sin_lat_b[1:] - sin_lat_b[0:-1])
    dx = np.abs(lon_b[1:] - lon_b[0:-1])

    dy2d, dx2d = np.meshgrid(dx, dy)

    area = (np.pi / 180.0) * (rad_earth**2) * dy2d * dx2d

    return area


def x_average_dataset(dset, region=None, lon_0=None):
    """Function to average a dataset along the x-dimension

    This function performs a zonal mean on an xarray-dataset

    Parameters
    ----------
    dset : xarray.Dataset
        Input dataset
    region : str, optional
        Basin/region to subset. Options are "atlarc" and "indpac",
        by default None (global)
    lon_0 : float, optional
        Slice along a specific meridian, by default None

    Returns
    -------
    xarray.Dataset
    """

    xcoords = dset.cf.coordinates["longitude"]
    xdims = [dset[x].dims[-1] for x in xcoords]

    ycoords = dset.cf.coordinates["latitude"]

    ydims = []
    for ycoord in ycoords:
        ydim = list(dset[ycoord].dims)

        if len(ydim) > 1:
            ydim = ydim[-2]
            multidim = True
        else:
            ydim = ydim[-1]
            multidim = False

        ydims.append(ydim)

    z_i = dset["z_i"] if "z_i" in dset.keys() else None

    if region is not None:
        result = xr.Dataset()
        basin = {}

        varlist = list(dset.keys()) + ycoords if multidim else list(dset.keys())

        for var in varlist:
            xcoord = dset[var].cf.coordinates["longitude"][0]
            xdim = list(dset[xcoord].dims)[-1]

            ycoord = dset[var].cf.coordinates["latitude"][0]
            ydim = list(dset[ycoord].dims)
            ydim = ydim[-2] if len(ydim) > 1 else ydim[-1]

            if xcoord == "geolon_u":
                wet = "wet_u"
            elif xcoord == "geolon_v":
                wet = "wet_v"
            elif xcoord == "geolon_c":
                wet = "wet_c"
            else:
                wet = "wet"

            dimset = (ydim, xdim)
            if dimset not in basin.keys():
                _dset = xr.Dataset()
                _dset[var] = dset[var]
                _codes = cmip_basins.generate_basin_codes(_dset, xcoord, ycoord, wet)
                regions = {
                    "atlarc": xr.where(
                        (_codes == 2) | (_codes == 4), 1.0, np.nan, keep_attrs=True
                    ),
                    "indpac": xr.where(
                        (_codes == 3) | (_codes == 5), 1.0, np.nan, keep_attrs=True
                    ),
                }
                basin[dimset] = regions
            else:
                regions = basin[dimset]

            result[var] = (dset[var] * regions[region]).mean(xdim)
            result[var].attrs = dset[var].attrs

        result = result.set_coords(ycoords)

    elif lon_0 is not None:
        result = xr.Dataset()
        varlist = list(dset.keys()) + ycoords if multidim else list(dset.keys())
        for var in varlist:
            result[var] = x_slice(dset[var], lon_0)
            result[var].attrs = dset[var].attrs

        result = result.set_coords(ycoords)

    else:
        if multidim:
            dset = dset.reset_coords(ycoords)
            dset = dset.mean(xdims, keep_attrs=True)
            result = dset.set_coords(ycoords)
        else:
            result = dset.mean(xdims, keep_attrs=True)

    mappings = {result[x].dims[0]: x for x in ycoords}

    result["z_i"] = z_i
    result.set_coords("z_i")

    return result


def verify_dim_lens(var1, var2, verbose=True):
    """Function to test the equality of dimension lengths

    This function determines if the shared dimensions between two
    data arrays are of equal length

    Parameters
    ----------
    var1 : xarray.DataArray
    var2 : xarray.DataArray
    verbose : bool, optional
        Issue warnings if dimensions do not agree, by default True

    Returns
    -------
    bool
    """

    dims = list(set(var1.dims).intersection(set(var2.dims)))
    exception_count = 0
    for dim in dims:
        try:
            assert len(var1[dim]) == len(var2[dim]), (
                f"Different {dim} lengths for {var1.name}: "
                + f"{len(var1[dim])}, {len(var2[dim])} "
                + "Consider symmetric vs. non-symmetric memory "
                + "output vs grid definition."
            )
        except AssertionError as exc:
            warnings.warn(str(exc))
            exception_count += 1
    return exception_count == 0



""" geoslice.py : toolkit for selecting a regions """

import itertools
import numpy as np

__all__ = [
    "get_indices_2dcurvilinear",
    "get_indices_1dlatlon",
    "get_indices",
    "geoslice",
    "x_slice",
]


def get_indices_2dcurvilinear(lat_grid, lon_grid, y, x):
    """This function returns the j,i indices for the grid point closest to the input lon(i,j),lat(i,j) coordinates."""
    """It returns the j,i indices."""

    cost = np.fabs((lon_grid - x) ** 2 + (lat_grid - y) ** 2)
    costm = np.where(cost == cost.min())
    j0, i0 = costm[0][0], costm[1][0]
    return j0, i0


def get_indices_1dlatlon(lat_grid, lon_grid, y, x):
    """This function returns the j,i indices for the grid point closest to the input lon(i),lat(j) coordinates."""
    """It returns the j,i indices."""

    lons = np.fabs(np.squeeze(lon_grid) - x)
    lonm = np.where(lons == lons.min())
    lats = np.fabs(np.squeeze(lat_grid) - y)
    latm = np.where(lats == lats.min())
    j0, i0 = latm[0][0], lonm[0][0]
    return j0, i0


def get_indices(lat_grid, lon_grid, y, x):
    """Returns j,i indices of lat_grid and lon grid nearest to x,y"""
    if len(lon_grid.shape) == 1:
        J, I = get_indices_1dlatlon(lat_grid, lon_grid, y, x)
    else:
        J, I = get_indices_2dcurvilinear(lat_grid, lon_grid, y, x)
    return J, I


def geoslice(arr, x=(None, None), y=(None, None), ycoord=None, xcoord=None):
    if xcoord is None:
        xcoord = arr.cf.coordinates["longitude"][0]
    if ycoord is None:
        ycoord = arr.cf.coordinates["latitude"][0]

    xdim = arr[xcoord].dims[-1]
    ydim = arr[ycoord].dims[-2] if len(arr[ycoord].dims) == 2 else arr[ycoord].dims[0]

    combinations = list(itertools.product(y, x))
    xlist = [x[1] for x in combinations if x[1] is not None]
    ylist = [y[0] for y in combinations if y[0] is not None]

    lower_left = (np.min(ylist), np.min(xlist))
    lower_right = (np.min(ylist), np.max(xlist))
    upper_left = (np.max(ylist), np.min(xlist))
    upper_right = (np.max(ylist), np.max(xlist))

    lower_left_ij = get_indices(arr[ycoord], arr[xcoord], *lower_left)
    lower_right_ij = get_indices(arr[ycoord], arr[xcoord], *lower_right)
    upper_left_ij = get_indices(arr[ycoord], arr[xcoord], *upper_left)
    upper_right_ij = get_indices(arr[ycoord], arr[xcoord], *upper_right)

    combinations = [lower_left_ij, lower_right_ij, upper_left_ij, upper_right_ij]
    xlist = [x[1] for x in combinations if x[1] is not None]
    ylist = [y[0] for y in combinations if y[0] is not None]
    xrng = (np.min(xlist), np.max(xlist))
    yrng = (np.min(ylist), np.max(ylist))

    return arr.isel({ydim: slice(*yrng), xdim: slice(*xrng)})


def x_slice(arr, lon_0, xcoord=None, ycoord=None):
    """Function to slice an array at a specific longitude

    Parameters
    ----------
    arr : xarray.core.DataArray
        Input data array
    lon_0 : float
        Longitude to perform slice
    xcoord : str, optional
        Name of x-coordinate otherwise it is inferred, by default None
    ycoord : str, optional
        Name of y-coordinate otherwise it is inferred, by default None

    Returns
    -------
    xarray.core.DataArray
        Sliced data array

    """
    if xcoord is None:
        xcoord = arr.cf.coordinates["longitude"][0]
    if ycoord is None:
        ycoord = arr.cf.coordinates["latitude"][0]

    xdim = arr[xcoord].dims[-1]

    j, i = get_indices(arr[ycoord], arr[xcoord], arr[ycoord].mean(), lon_0)

    result = arr.isel({xdim: i})

    return result



#!/usr/bin/env python

import argparse
import warnings
import sys
import xarray as xr
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import momgrid

warnings.filterwarnings(
    "ignore",
    message="The input coordinates to pcolormesh are interpreted as cell centers, but are not monotonically increasing or decreasing.*",
)


def main(args):
    resolution = 1.0
    plot_type = "yx"
    level = 0.0
    region = None
    tdim = "time"
    projection = None
    # projection = "Robinson"

    var = args.varname[0]
    varlist = args.varname

    ds1 = momgrid.Gridset(args.ds1)
    ds2 = momgrid.Gridset(args.ds2)

    ds1.subset(varlist)
    ds2.subset(varlist)

    # If an xy plot is requested, extract the desired depth level
    if plot_type == "yx":
        try:
            ds1.data = momgrid.util.extract_level(ds1.data, level)
            ds2.data = momgrid.util.extract_level(ds2.data, level)
        except:
            pass

    if tdim in ds1.data.dims:
        ds1.data = ds1.data.mean(tdim, keep_attrs=True)

    if tdim in ds2.data.dims:
        ds2.data = ds2.data.mean(tdim, keep_attrs=True)

    # Check if grids are identical and regrid if necessary
    if ds1.model == ds2.model:
        ds1 = ds1.data
        ds2 = ds2.data
    else:
        ds1 = ds1.regrid(resolution=resolution)
        ds2 = ds2.regrid(resolution=resolution)

    if plot_type == "yz":
        ds1 = momgrid.util.x_average_dataset(ds1, region=region)
        ds2 = momgrid.util.x_average_dataset(ds2, region=region)

    projection = ccrs.__dict__[projection]() if projection is not None else None

    _ = momgrid.plot.compare_2d(
        ds1[var],
        ds2[var],
        dpi=90,
        singlepanel=False,
        plot_type=plot_type,
        projection=projection,
    )

    plt.show()


def list_vars(args):
    ds1 = xr.open_mfdataset(args.ds1, decode_times=False)
    ds2 = xr.open_mfdataset(args.ds2, decode_times=False)
    varlist = set(ds1.keys()).intersection(ds2.keys())
    for x in sorted(varlist):
        _longname = (
            ds1[x].attrs["long_name"] if "long_name" in ds1[x].attrs.keys() else ""
        )
        print(f" *  {x}  : {_longname}")


if __name__ == "__main__":
    # Create the parser
    parser = argparse.ArgumentParser(
        description="Compare two datasets generated by the MOM ocean model."
    )

    # Add arguments
    parser.add_argument(
        "ds1", type=str, help="Comma-separated list of filenames for the first dataset."
    )
    parser.add_argument(
        "ds2",
        type=str,
        help="Comma-separated list of filenames for the second dataset.",
    )
    parser.add_argument(
        "varname",
        type=str,
        nargs="?",
        default=None,
        help="Variable name to compare. If omitted, a list of available variables will be offered",
    )

    # Parse arguments
    args = parser.parse_args()

    args.ds1 = args.ds1.split(",")
    args.ds2 = args.ds2.split(",")

    if args.varname is None:
        list_vars(args)
        sys.exit(1)
    else:
        args.varname = args.varname.split(",")

    main(args)

    sys.exit(0)



""" external.py: functions to interface with external packages """

import warnings
import xarray as xr
import xesmf as xe
import pkg_resources as pkgr
from momgrid.util import is_symmetric

__all__ = ["build_regridder_weights", "static_to_xesmf", "woa18_grid"]


def build_regridder_weights(src, dst, periodic=True):
    """Function to generate pre-calculated xesmf weight files"""

    methods = ["bilinear", "nearest_s2d", "nearest_d2s"]

    if (
        set(["lat_b", "lon_b"]).issubset(
            list(src.keys()) + list(src.coords) + list(src.dims)
        )
    ) and (
        set(["lat_b", "lon_b"]).issubset(
            list(dst.keys()) + list(dst.coords) + list(dst.dims)
        )
    ):
        bounds = True
        methods = methods + ["conservative", "conservative_normed", "patch"]
    else:
        bounds = False

    files = []
    for method in methods:
        _ = xe.Regridder(src, dst, method, periodic=periodic)
        _ = _.to_netcdf()
        files.append(_)

    return files


def static_to_xesmf(dset, grid_type="t", filename=None):
    """Function to convert a MOM6 static file to one that can be
    fed into xesmf routines.

    Parameters
    ----------
    dset : xarray.Dataset
        MOM6 static file dataset
    grid_type : str
        Grid type (t,u,v,c), optional. By default "t"

    Returns
    -------
    xarray.Dataset
        Xarray dataset to compatible with xesmf
    """

    assert isinstance(dset, xr.Dataset), "Input must be an xarray dataset."

    if grid_type == "t":
        dsout = xr.Dataset(
            {
                "lat": dset.geolat,
                "lon": dset.geolon,
                "mask": dset.wet,
            }
        )

        if is_symmetric(dset):
            dsout["lat_b"] = dset.geolat_c
            dsout["lon_b"] = dset.geolon_c
        else:
            warnings.warn("Grid is not symmetric, skipping bounds")

    elif grid_type == "u":
        dsout = xr.Dataset(
            {
                "lat": dset.geolat_u,
                "lon": dset.geolon_u,
            }
        )

        if "wet_u" in dset.keys():
            dsout["mask"] = dset.wet_u
        else:
            warnings.warn("Wet mask not present.")

    elif grid_type == "v":
        dsout = xr.Dataset(
            {
                "lat": dset.geolat_v,
                "lon": dset.geolon_v,
            }
        )

        if "wet_v" in dset.keys():
            dsout["mask"] = dset.wet_v
        else:
            warnings.warn("Wet mask not present.")

    elif grid_type == "c":
        dsout = xr.Dataset(
            {
                "lat": dset.geolat_c,
                "lon": dset.geolon_c,
            }
        )

        if "wet_c" in dset.keys():
            dsout["mask"] = dset.wet_c
        else:
            warnings.warn("Wet mask not present.")

    else:
        raise ValueError(f"Unsupported grid type: {grid_type}")

    dsout = dsout.reset_coords(drop=True)

    if filename is not None:
        filename = str(filename)
        dsout.to_netcdf(filename)
        result = filename
    else:
        result = dsout

    return result


def woa18_grid(resolution=0.25):
    """Function to return World Ocean Atlas horizontal grid

    Parameters
    ----------
    resolution : float, optional
        Horizontal resolution (0.25 or 1.0), by deafult 0.25

    Returns
    -------
    xarray.Dataset
    """

    if resolution == 0.25:
        res_str = "025"
    elif resolution == 1.0:
        res_str = "1"
    else:
        raise ValueError(
            f"Unknown resolution: {resolution}. Must be either 0.25 or 1.0"
        )

    fpath = pkgr.resource_filename("momgrid", f"grids/WOA18_{res_str}deg_horiz_grid.nc")

    dset = xr.open_dataset(fpath)

    return dset
